/*
 * JQuery boundsliders plugin
 * Version 1.0.1 (27/03/2010)
 *
 * Home page: http://projects.nocternity.net/index.py/en/jquery-boundsliders
 * Examples: http://projects.nocternity.net/index.py/en/jqbs-demo
 *
 * Copyright (c) 2010 Emmanuel Benoît
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

$(function(){

	// Default values for the various options
	var _defaults = {
		'totalValue':		100 ,
		'jquerySliders':	false ,
		'bigIncrement':		null ,
		'buttons':		[ ] ,
		'lock':			null ,
		'buttonsLocation':	'around'
	};

	// Number verification function
	var _checkNumber = function ( _value , _minValue , _name ) {
		if ( typeof _value != 'number' ) {
			throw new Error( "$.fn.boundSliders(): " + _name + " option should be a number" );
		}
		if ( _value < _minValue ) {
			throw new Error( "$.fn.boundSliders(): " + _name + " option should be at least " + _minValue );
		}
		if ( _value % 1 != 0 ) {
			throw new Error( "$.fn.boundSliders(): " + _name + " option should be an integer" );
		}
	};

	// Function that verifies options
	var _checkOptions = function ( _options ) {
		// Check total value
		_checkNumber( _options.totalValue , 1 , 'totalValue' );

		// Check big increment, if present
		if ( _options.bigIncrement !== null ) {
			_checkNumber( _options.bigIncrement , 2 , 'bigIncrement' );
		}

		// Check buttons
		if ( ! $.isArray( _options.buttons ) ) {
			throw new Error( "$.fn.boundSliders(): buttons option should be an array" );
		}
		if ( _options.buttons.length < 4 ) {
			var _nButtons = [ null , null , null , null ];
			for ( var i in _options.buttons ) {
				_nButtons[ i ] = _options.buttons[ i ];
			}
			_options.buttons = _nButtons;
		}
		for ( var i in _options.buttons ) {
			if ( _options.buttons[ i ] !== null && typeof _options.buttons[ i ] != 'string' ) {
				throw new Error( "$.fn.boundSliders(): buttons[" + i + "] is not a string" );
			}
		}

		// Check lock
		if ( _options.lock !== null ) {
			if ( ! $.isArray( _options.lock ) ) {
				throw new Error( "$.fn.boundSliders(): lock option should be an array" );
			}
			if ( _options.lock.length != 2 ) {
				throw new Error( "$.fn.boundSliders(): lock option should have 2 elements" );
			}
			for ( var i in _options.lock ) {
				if ( typeof _options.lock[ i ] != 'string' ) {
					throw new Error( "$.fn.boundSliders(): lock[" + i + "] is not a string" );
				}
			}
		}

		// Check location
		if ( _options.buttonsLocation != 'before' && _options.buttonsLocation != 'around' && _options.buttonsLocation != 'after' ) {
			throw new Error( "$.fn.boundSliders(): buttonsLocation option should be 'before', 'around' or 'after'" );
		}
	};

	// Reads all fields and initialises bound sliders data
	var _getInitialValues = function( _parent , _data ) {
		var _sum = 0;
		var _count = 0;

		$( 'input[type=text]' , _parent ).each(function() {
			var _name = $(this).attr('name');
			var _value = $(this).attr('value');

			_value = parseInt( _value , 10 );
			if ( _value === Number.NaN ) {
				throw new Error( "$.fn.boundSliders(): field '" + _name + "' has a non-numeric value" );
			}
			_data[ _name ] = {
				'locked':	false ,
				'value':	_value
			};
			_sum += _value;
			_count ++;
		});

		if ( _count < 3 ) {
			throw new Error( "$.fn.boundSliders(): at least 3 fields required" );
		}
		_data._count = _count;

		return _sum;
	};

	// Button creator
	var _createButton = function ( _text , _image ) {
		var _button = $( '<button/>' ).addClass('slider-button').attr({
			'title':	_text
		});

		if ( _image ) {
			_button.addClass( 'image' ).append( $('<img/>').attr({ 'src' : _image , 'alt' : _text }) );
		} else {
			_button.append( _text );
		}

		return _button;
	};

	// Bound sliders "class"
	var _BoundSliders = function ( _parent , _options ) {
		// Get initial values
		var _data = { };
		if ( _getInitialValues( _parent , _data ) != _options.totalValue ) {
			var _mod = _options.totalValue % _data._count;
			var _div = ( _options.totalValue - _mod ) / _data._count;
			for ( var i in _data ) {
				if ( i == "_count" ) {
					continue;
				}
				_data[ i ].value = _div + ( _mod ? 1 : 0 );
				_mod -= ( _mod ? 1 : 0 );
			}
		}

		var _locked = 0;
		var _movingTotal = _options.totalValue;

		// Internal function that disables lock buttons if all but two locks are engaged
		var _updateLocks = function( ) {
			var _disabled = ( _locked == _data._count - 2 );
			for ( var i in _data ) {
				if ( i != "_count" && ! _data[ i ].locked ) {
					if ( _disabled ) {
						_data[ i ].lock.attr( 'disabled' , true );
					} else {
						_data[ i ].lock.removeAttr( 'disabled' );
					}
				}
			}
		};

		// Internal function that updates field values
		var _updateFields = function( ) {
			var _disabled = ( _locked == _data._count - 2 );
			for ( var i in _data ) {
				if ( i != "_count" ) {
					_data[ i ].field.attr( 'value' , _data[i].value );
					if ( _options.jquerySliders ) {
						_data[ i ].jqSlider.slider( "value" , _data[i].value );
					}
				}
			}
		};

		// Internal function that changes one of the values
		var _changeValue = function ( _name , _change ) {
			var _entry = _data[ _name ];
			var _total = _entry.locked ? _options.totalValue : _movingTotal;

			// Adjust value to prevent going under 0 or over max
			var _oChange = _change;
			if ( _change < 0 && _change + _entry.value < 0 ) {
				_change = - _entry.value;
			} else if ( _change > 0 && _change + _entry.value > _total ) {
				_change = _total - _entry.value;
			}
			if ( _entry.locked && _change > _movingTotal ) {
				_change = _movingTotal;
			}
			if ( _change == 0 || _options.jquerySliders && _oChange != _change ) {
				return false;
			}
			_entry.value += _change;
			if ( _entry.locked ) {
				_movingTotal -= _change;
			}

			// Find all fields that need to be modified
			var _toChange = [];
			var _inc = - _change / Math.abs( _change );
			for ( var i in _data ) {
				if ( i == "_count" || i == _name || _data[ i ].locked ) {
					continue;
				}
				_toChange.push( _data[ i ] );
			}

			// Modify all fields
			while ( _change != 0 ) {
				var _toIncrease = _toChange[ 0 ];
				var _xValue = _toIncrease.value;

				for ( var i = 1 ; i < _toChange.length ; i ++ ) {
					var _tc = _toChange[i];
					if ( ( _inc < 0 && _tc.value == 0 ) || ( _inc > 0 && _tc.value == _movingTotal ) ) {
						continue;
					}
					if ( ( _inc < 0 && _xValue < _tc.value ) || ( _inc > 0 && _xValue > _tc.value ) ) {
						_toIncrease = _tc;
						_xValue = _toIncrease.value;
					}
				}
				_toIncrease.value += _inc;
				_change += _inc;
			}

			_updateFields( );
			return true;
		};

		// Internal function that locks / unlocks a value
		var _toggleLock = function ( _name ) {
			var _entry = _data[ _name ];

			if ( _entry.locked ) {
				_entry.locked = false;
				_movingTotal += _entry.value;
				_locked --;
			} else {
				_entry.locked = true;
				_movingTotal -= _entry.value;
				_locked ++;
			}

			if ( _options.lock ) {
				$( 'img' , _entry.lock ).attr( 'src' , _options.lock[ _entry.locked ? 0 : 1 ] );
			}

			_updateLocks( );
		};

		// Create controls
		$( 'input[type=text]' , _parent ).each(function() {
			var _name = $(this).attr( 'name' );
			var _decrease , _increase , _lock;
			$(this).attr( 'readonly' , true );

			if ( _options.jquerySliders ) {
				_increase = $('<span/>').addClass( 'boundsliders-slider' ).css( 'display' , 'inline-block' );
				if ( _options.buttonsLocation == 'after' ) {
					_increase.insertAfter( $(this) );
				} else {
					_increase.insertBefore( $(this) );
				}
				_increase.slider({
					'max':		_options.totalValue ,
					'value':	_data[ _name ].value ,
					'slide':	function( evt , ui ) {
						return _changeValue( _name , ui.value - _data[ _name ].value );
					}
				});

				_data[ _name ].jqSlider = _increase;
			} else {
				// Create '-' button
				_decrease = _createButton( '-' , _options.buttons[1] ).click( function() {
					_changeValue( _name , -1 );
					return false;
				});
				if ( _options.buttonsLocation == 'after' ) {
					_decrease.insertAfter( $(this) );
				} else {
					_decrease.insertBefore( $(this) );
				}

				// Create '+' button
				_increase = _createButton( '+' , _options.buttons[0] ).click( function() {
					_changeValue( _name , 1 );
					return false;
				});
				if ( _options.buttonsLocation == 'around' ) {
					_increase.insertAfter( $(this) );
				} else {
					_increase.insertAfter( _decrease );
				}
			}

			// Create lock button
			if ( _options.lock ) {
				_lock = _createButton( 'lock' , _options.lock[1] );
			} else {
				_lock = $( '<input/>' ).addClass( 'slider-lock' ).attr({
					'type':		'checkbox' ,
					'name':		_name + '-lock' ,
					'title':	'lock'
				});
			}
			_lock.click( function() {
				_toggleLock( _name );
				return ( _options.lock === null );
			}).insertAfter( _increase );

			// Create "big increment" buttons
			if ( !_options.jquerySliders &&_options.bigIncrement ) {
				_createButton( '++' , _options.buttons[2] ).click( function() {
					_changeValue( _name , _options.bigIncrement );
					return false;
				}).insertAfter( _increase );
				_createButton( '--' , _options.buttons[3] ).click( function() {
					_changeValue( _name , -_options.bigIncrement );
					return false;
				}).insertBefore( _decrease );
			}

			_data[ _name ].lock = _lock;
			_data[ _name ].field = $(this);
		});

		_updateFields( );
	};

	// JQuery function
	$.fn.boundSliders = function ( _options ) {
		if ( !_options ) {
			_options = {};
		}
		var _rOptions = $.extend( true , {} , _defaults , _options );
		_checkOptions( _rOptions );
		return this.each( function() {
			var _value = $(this).data( 'boundSliders' );
			if ( ! _value ) {
				$(this).data( 'boundSliders' , new _BoundSliders( this , _rOptions ) );
			}
		});
	};

});
